"""Ozone specific statistic calculation routines.

This file comprises all functions that are needed to calculate the ozone
metrics of (time series) data.

Note: These statistical functions should be applied to time series data
and the time values should be given as ``solar time`` or ``local time``.

This module contains the following functions:
max1h_values - calculate the daily maximum hourly value
dma8epa - calculate the daily maximum 8-hour running mean
dma8epax - calculate the daily maximum 8-hour running mean
dma8eu - calculate the daily maximum 8-hour running mean
dma8epa_strict - calculate the daily maximum 8-hour running mean
dma8epax_strict - calculate the daily maximum 8-hour running mean
dma8eu_strict - calculate the daily maximum 8-hour running mean
avgdma8epax - calculate the average daily maximum 8-hour running mean
drmdmax1h - calculate the running mean of the 3-months average
somo10 - calculate the SOMO10 metric
somo35 - calculate the SOMO35 metric
somo10_strict - calculate the SOMO10 metric
somo35_strict - calculate the SOMO35 metric
w90 - calculate the W90 metric
aot40 - calculate the AOT40 metric
daylight_aot40 - calculate the daylight AOT40 metric
dark_aot40 - calculate the "dark" AOT40 metric
w126 - calculate the W126 metric
w126_24h - calculate the W126 metric over the 24 hour period of a day
nvgt050 - calculate the NVGT050 metric
nvgt060 - calculate the NVGT060 metric
nvgt070 - calculate the NVGT070 metric
nvgt080 - calculate the NVGT080 metric
nvgt090 - calculate the NVGT090 metric
nvgt100 - calculate the NVGT100 metric
nvgt120 - calculate the NVGT120 metric
nvgtall - calculate all nvgt metrics together
"""

import numpy as np
import pandas as pd

from toarstats.metrics.stats_utils import (
    aot40stat, dma8_processor, get_elevation_angle, kth_highest, prepare_data,
    register, resample_with_date, somo_processor, stat_processor_1,
    threshold_processor, w126stat
)


@register("max1h")
def max1h_values(ser, ref, mtype, metadata, seasons, min_data_capture):
    """Calculate the daily maximum hourly value.

    A day is defined from 0-23h ``solar time``. ``DMAX1H`` is computed
    under the condition that at least the fraction min data capture of
    hourly values in a day are present. Default min data capture is 75%
    (i.e. 18 hours out of 24). Otherwise the daily value will be coded
    as missing value.

    :param ser: a series containing the values
    :param ref: reference series
    :param mtype: aggregation type
    :param metadata: metadata needed for some statistics
    :param seasons: season names
    :param min_data_capture: a fractional coverage value

    :return: A list of dictionaries containing the processed data
    """
    return stat_processor_1("max1h", ser, ref, mtype, seasons, how="max",
                            minfrac=min_data_capture)


@register("dma8epa")
def dma8epa(ser, ref, mtype, metadata, seasons, min_data_capture):
    """Calculate the daily maximum 8-hour running mean (EPA).

    Daily maximum 8-hour running mean value according to US EPA. 8-hour
    averages are calculated for 24 bins starting at 0h ``local time``.
    The 8-hour running mean for a particular hour is calculated on the
    concentration for that hour plus the following 7 hours. If with the
    default min data capture criterion less than 75% of data are present
    (i.e. less than 6 hours), the average is considered missing.

    In the original EPA definition, a day is only valid if at least 18
    out of the 24 8-hour averages are valid. This criterion is not
    applied here. Use ``dma8epa_strict`` if you need it.

    :param ser: a series containing the values
    :param ref: reference series
    :param mtype: aggregation type
    :param metadata: metadata needed for some statistics
    :param seasons: season names
    :param min_data_capture: a fractional coverage value

    :return: A list of dictionaries containing the processed data
    """
    return dma8_processor(ser, ref, mtype, metadata, seasons, min_data_capture,
                          tshift=-7, dname="EPA")


@register("dma8epax")
def dma8epax(ser, ref, mtype, metadata, seasons, min_data_capture):
    """Calculate the daily maximum 8-hour running mean (new EPA).

    Daily maximum 8-hour running mean value according to newly proposed
    US EPA standard. 8-hour averages are calculated for 17 bins starting
    at 7h ``local time`` with the last interval starting at 23h. The
    8-hour running mean for a particular hour is calculated on the
    concentration for that hour plus the following 7 hours. If with the
    default min data capture criterion less than 75% of data are
    present, the average is considered missing.

    In the original EPA definition, the maximum 8-hour average value of
    each day is only saved if at least 13 out of the 17 8-hour averages
    are valid. This criterion is not applied here. Use
    ``dma8epax_strict`` if you need it.

    :param ser: a series containing the values
    :param ref: reference series
    :param mtype: aggregation type
    :param metadata: metadata needed for some statistics
    :param seasons: season names
    :param min_data_capture: a fractional coverage value

    :return: A list of dictionaries containing the processed data
    """
    return dma8_processor(ser, ref, mtype, metadata, seasons, min_data_capture,
                          tshift=-7, hours=[7, 23], dname="EPAx")


@register("dma8eu")
def dma8eu(ser, ref, mtype, metadata, seasons, min_data_capture):
    """Calculate the daily maximum 8-hour running mean (EU).

    Daily maximum 8-hour running mean value according to EU. 8-hour
    averages are calculated for 24 bins starting at 17h ``local time``
    of the previous day. The 8-hour running mean for a particular hour
    is calculated on the concentration for that hour plus the following
    7 hours. If with the default min data capture criterion less than
    75% of data are present, the average is considered missing.

    In the original EU definition, the maximum 8-hour average value of
    each day is only saved if at least 18 out of the 24 8-hour averages
    are valid. This criterion is not applied here. Use ``dma8eu_strict``
    if you need it.

    :param ser: a series containing the values
    :param ref: reference series
    :param mtype: aggregation type
    :param metadata: metadata needed for some statistics
    :param seasons: season names
    :param min_data_capture: a fractional coverage value

    :return: A list of dictionaries containing the processed data
    """
    return dma8_processor(ser, ref, mtype, metadata, seasons, min_data_capture,
                          dname="EU")


@register("dma8epa_strict")
def dma8epa_strict(ser, ref, mtype, metadata, seasons, min_data_capture):
    """Calculate the daily maximum 8-hour running mean (EPA).

    As ``dma8epa``, but with strict min data capture, i.e. requiring at
    least 18 valid 8-hour averages in a day.

    :param ser: a series containing the values
    :param ref: reference series
    :param mtype: aggregation type
    :param metadata: metadata needed for some statistics
    :param seasons: season names
    :param min_data_capture: a fractional coverage value

    :return: A list of dictionaries containing the processed data
    """
    return dma8_processor(ser, ref, mtype, metadata, seasons, min_data_capture,
                          tshift=-7, dname="EPA_strict", strict=True)


@register("dma8epax_strict")
def dma8epax_strict(ser, ref, mtype, metadata, seasons, min_data_capture):
    """Calculate the daily maximum 8-hour running mean (new EPA).

    As ``dma8epax``, but with strict min data capture, i.e. requiring at
    least 13 valid 8-hour averages in a day.

    :param ser: a series containing the values
    :param ref: reference series
    :param mtype: aggregation type
    :param metadata: metadata needed for some statistics
    :param seasons: season names
    :param min_data_capture: a fractional coverage value

    :return: A list of dictionaries containing the processed data
    """
    return dma8_processor(ser, ref, mtype, metadata, seasons, min_data_capture,
                          tshift=-7, hours=[7, 23], dname="EPAx_strict",
                          strict=True)


@register("dma8eu_strict")
def dma8eu_strict(ser, ref, mtype, metadata, seasons, min_data_capture):
    """Calculate the daily maximum 8-hour running mean (EU).

    As ``dma8eu``, but with strict min data capture, i.e. requiring at
    least 18 valid 8-hour averages in a day.

    :param ser: a series containing the values
    :param ref: reference series
    :param mtype: aggregation type
    :param metadata: metadata needed for some statistics
    :param seasons: season names
    :param min_data_capture: a fractional coverage value

    :return: A list of dictionaries containing the processed data
    """
    return dma8_processor(ser, ref, mtype, metadata, seasons, min_data_capture,
                          dname="EU_strict", strict=True)


@register("avgdma8epax")
def avgdma8epax(ser, ref, mtype, metadata, seasons, min_data_capture):
    """Calculate the average daily maximum 8-hour running mean.

    As ``dma8epax``, but calculate the average daily ``DMA8`` value
    instead of the k-th highest (new GBD metrics).

    :param ser: a series containing the values
    :param ref: reference series
    :param mtype: aggregation type
    :param metadata: metadata needed for some statistics
    :param seasons: season names
    :param min_data_capture: a fractional coverage value

    :return: A list of dictionaries containing the processed data
    """
    return dma8_processor(ser, ref, mtype, metadata, seasons, min_data_capture,
                          tshift=-7, hours=[7, 23], dname="EPAx", average=True)


@register("drmdmax1h")
def drmdmax1h(ser, ref, mtype, metadata, seasons, min_data_capture):
    """Calculate the running mean of the 3-months average of ``DMAX1H``.

    Starting from the first day of the data record, then the second day,
    and so on, 3-month averages will be calculated if with the default
    min data capture criterion 75% of days in the 3-months interval of
    fixed 90-days length have a valid ``DMAX1H`` value.

    Note that this metric is calculated differently than others. The
    running mean is evaluated on the ``ref`` time series (with filled
    time slots), and the split into seasons (if necessary) happens after
    the computation of the daily values.

    :param ser: a series containing the values
    :param ref: reference series
    :param mtype: aggregation type
    :param metadata: metadata needed for some statistics
    :param seasons: season names
    :param min_data_capture: a fractional coverage value

    :raises ValueError: raised if ``mtype = "monthly"``

    :return: A list of dictionaries containing the processed data
    """
    mincount = int(min_data_capture * 90)
    tmp1 = max1h_values(
        ser, ref, "daily", metadata, seasons, min_data_capture
    )[0]
    tref = pd.date_range(
        start=tmp1["ser"].index.min(), end=tmp1["ser"].index.max(), freq="D"
    )
    tmp1ref = tmp1["ser"].reindex(tref)
    # Calculate 3-months rolling average (shift by 44 days to store
    # value at an appropriate time index).
    tmp2 = tmp1ref.rolling(window=90, min_periods=mincount).mean().shift(-44)
    tmp2 = tmp2.reindex(tmp1["ser"].index)
    if mtype == "daily":
        tmp1["ser"] = tmp2
        res = [tmp1]
    elif mtype == "monthly":
        raise ValueError("Invalid mtype")  # pragma: no cover
    else:
        tmpres = stat_processor_1("drmdmax1h", tmp2, tmp1ref, mtype, seasons,
                                  func=resample_with_date)
        res = []
        for r in tmpres:
            r1 = r.copy()
            r1["ser"] = r["ser"]["values"]
            r2 = r.copy()
            r2["ser"] = r["ser"]["date_of_maximum"].dt.day_of_year.astype(int)
            r2["name"] = f"day_of_{r2['name']}"
            res.extend([r1, r2])
    return res


@register("somo10")
def somo10(ser, ref, mtype, metadata, seasons, min_data_capture):
    """Calculate the SOMO10 metric.

    Sum of excess of daily maximum 8-hour means (EU Airbase standard)
    over the cut-off of 10 ppb, i.e. 20 ug/m3 calculated for all days in
    a period. ``SOMO10`` will be set to missing if with the default data
    capture criterion less than 75% of days are available. The quantity
    will be weighted by the number of theoretical days over the number
    of available days.

    :param ser: a series containing the values
    :param ref: reference series
    :param mtype: aggregation type
    :param metadata: metadata needed for some statistics
    :param seasons: season names
    :param min_data_capture: a fractional coverage value

    :return: A list of dictionaries containing the processed data
    """
    return somo_processor(ser, ref, mtype, metadata, seasons, min_data_capture,
                          10.)


@register("somo35")
def somo35(ser, ref, mtype, metadata, seasons, min_data_capture):
    """Calculate the SOMO35 metric.

    Sum of excess of daily maximum 8-hour means (EU Airbase standard)
    over the cut-off of 35 ppb, i.e. 70 ug/m3 calculated for all days in
    a period. ``SOMO35`` will be set to missing if with the default data
    capture criterion less than 75% of days are available. The quantity
    will be weighted by the number of theoretical days over the number
    of available days.

    :param ser: a series containing the values
    :param ref: reference series
    :param mtype: aggregation type
    :param metadata: metadata needed for some statistics
    :param seasons: season names
    :param min_data_capture: a fractional coverage value

    :return: A list of dictionaries containing the processed data
    """
    return somo_processor(ser, ref, mtype, metadata, seasons, min_data_capture,
                          35.)


@register("somo10_strict")
def somo10_strict(ser, ref, mtype, metadata, seasons, min_data_capture):
    """Calculate the SOMO10 metric.

    Sum of excess of daily maximum 8-hour means (EU Airbase standard)
    over the cut-off of 10 ppb, i.e. 20 ug/m3 calculated for all days in
    a period. ``SOMO10`` will be set to missing if with the default data
    capture criterion less than 75% of days are available. The quantity
    will be weighted by the number of theoretical days over the number
    of available days.

    :param ser: a series containing the values
    :param ref: reference series
    :param mtype: aggregation type
    :param metadata: metadata needed for some statistics
    :param seasons: season names
    :param min_data_capture: a fractional coverage value

    :return: A list of dictionaries containing the processed data
    """
    return somo_processor(ser, ref, mtype, metadata, seasons, min_data_capture,
                          10., strict=True)


@register("somo35_strict")
def somo35_strict(ser, ref, mtype, metadata, seasons, min_data_capture):
    """Calculate the SOMO35 metric.

    Sum of excess of daily maximum 8-hour means (EU Airbase standard)
    over the cut-off of 35 ppb, i.e. 70 ug/m3 calculated for all days in
    a period. ``SOMO35`` will be set to missing if with the default data
    capture criterion less than 75% of days are available. The quantity
    will be weighted by the number of theoretical days over the number
    of available days.

    :param ser: a series containing the values
    :param ref: reference series
    :param mtype: aggregation type
    :param metadata: metadata needed for some statistics
    :param seasons: season names
    :param min_data_capture: a fractional coverage value

    :return: A list of dictionaries containing the processed data
    """
    return somo_processor(ser, ref, mtype, metadata, seasons, min_data_capture,
                          35., strict=True)


@register("w90")
def w90(ser, ref, mtype, metadata, seasons, min_data_capture):
    """Calculate the W90 metric.

    Daily maximum ``W90`` 5-h Experimental Exposure Index:
    EI = sum( wi*Ci) with weight wi = 1/[1 +  M*exp(-A*Ci/1000)], where
    M = 1400, A = 90, and where Ci is the hourly average ``O3`` mixing
    ratio in units of ppb
    (Lefohn et al. (2010), Inhalation Toxicology, 22(12): 999-1011). For
    each day, 24 ``W90`` indices are computed as 5-hour sums, requiring
    that at least 4 of the 5 hours are valid data. If a sample consists
    of only 4 data points, a fifth value shall be constructed from
    averaging the 4 valid mixing ratios. The daily value is considered
    missing if with the default min data capture criterion less than 18
    ``W90`` values are valid. The annual (and also summer, seasonal, or
    monthly) statistics first calculate the daily ``W90`` indices and
    then returns the 4th highest daily ``W90`` index in the requested
    interval. This value will be coded as missing if with the default
    min data capture criterion less than 75% of days are available.

    :param ser: a series containing the values
    :param ref: reference series
    :param mtype: aggregation type
    :param metadata: metadata needed for some statistics
    :param seasons: season names
    :param min_data_capture: a fractional coverage value

    :return: A list of dictionaries containing the processed data
    """
    res = prepare_data(ser, ref, mtype, seasons, "w90")
    for r in res:
        M = 1400.
        A = 90.
        tmp1 = r["ser"]/(1.+M*np.exp(-A*ser/1000.))
        tmp2 = (tmp1.rolling(5, min_periods=4).sum().shift(-4)
                + (5-tmp1.rolling(5, min_periods=4).count().shift(-4))
                * tmp1.rolling(5, min_periods=4).mean().shift(-4))
        dcap = int(min_data_capture * 24)
        tmp2_resample = tmp2.resample("D")
        ser_res = tmp2_resample.max()
        count = tmp2_resample.count()
        ser_res[count < dcap] = np.nan
        if mtype == "daily":
            r["ser"] = ser_res
        else:
            ser_ref = r["ref"].resample("D").mean().dropna()
            r["ser"] = kth_highest(ser_res, ser_ref, mtype, 4,
                                   min_data_capture)
    return res


@register("aot40")
def aot40(ser, ref, mtype, metadata, seasons, min_data_capture):
    """Calculate the AOT40 metric.

    Daily 12-hour ``AOT40`` value computed over the 08:00-19:59h
    ``solar time`` interval (end points included). ``AOT40`` is defined
    as cumulative ozone above 40 ppb. If with the default min data
    capture criterion less than 75% of hourly values (i.e. less than 9
    out of 12 hours) are present, the daily value is considered missing.
    When there exist 75% or greater min data capture in the daily
    12-hour window, the scaling by fractional min data capture
    (ntotal/nvalid) will be utilized. Monthly, seasonal, summer, and
    annual statistics are generated by adding up all daily values within
    the requested interval. Missing days are added by replacing them
    with the mean daily ``AOT40`` value. If with the default min data
    capture criterion less than 75% of days are valid, the ``AOT40``
    value is considered missing.

    Note: when calculating ``seasonal`` statistics, the default seasons
    should be replaced by the wheat and/or rice growing season of the
    station location.

    :param ser: a series containing the values
    :param ref: reference series
    :param mtype: aggregation type
    :param metadata: metadata needed for some statistics
    :param seasons: season names
    :param min_data_capture: a fractional coverage value

    :return: A list of dictionaries containing the processed data
    """
    hour = ser.index.hour
    day = (hour >= 8) & (hour <= 19)
    hourref = ref.index.hour
    dayref = (hourref >= 8) & (hourref <= 19)
    return stat_processor_1("aot40", ser[day], ref[dayref], mtype, seasons,
                            func=aot40stat, how="sum",
                            minfrac=min_data_capture, threshold=40.)


@register("daylight_aot40")
def daylight_aot40(ser, ref, mtype, metadata, seasons, min_data_capture):
    """Calculate the daylight AOT40 metric.

    ``AOT40`` value computed during the period when solar elevation is
    > 5 degrees. See ``aot40`` for definition.

    :param ser: a series containing the values
    :param ref: reference series
    :param mtype: aggregation type
    :param metadata: metadata needed for some statistics
    :param seasons: season names
    :param min_data_capture: a fractional coverage value

    :return: A list of dictionaries containing the processed data
    """
    elref = get_elevation_angle(metadata.station_lat, metadata.station_lon,
                                ref.index)
    el = pd.Series(elref, ref.index).reindex(ser.index)
    return stat_processor_1("daylight_aot40", ser.where(el > 5.).dropna(),
                            ref.where(elref > 5.), mtype, seasons,
                            func=aot40stat, how="sum",
                            minfrac=min_data_capture, threshold=40.)


@register("dark_aot40")
def dark_aot40(ser, ref, mtype, metadata, seasons, min_data_capture):
    """Calculate the "dark" AOT40 metric.

    ``AOT40`` value computed during the period when solar elevation is
    <= 5 degrees. See ``aot40`` for definition.

    :param ser: a series containing the values
    :param ref: reference series
    :param mtype: aggregation type
    :param metadata: metadata needed for some statistics
    :param seasons: season names
    :param min_data_capture: a fractional coverage value

    :return: A list of dictionaries containing the processed data
    """
    elref = get_elevation_angle(metadata.station_lat, metadata.station_lon,
                                ref.index)
    el = pd.Series(elref, ref.index).reindex(ser.index)
    return stat_processor_1("dark_aot40", ser.where(el <= 5.).dropna(),
                            ref.where(elref <= 5.), mtype, seasons,
                            func=aot40stat, how="sum",
                            minfrac=min_data_capture, threshold=40.)


@register("w126")
def w126(ser, ref, mtype, metadata, seasons, min_data_capture):
    """Calculate the W126 metric.

    Daily ``W126`` index over the 8:00-19:59h ``solar time`` interval
    (see http://www.epa.gov/ttn/analysis/w126.htm).
    W126 = sum( wi*Ci) with weight wi = 1/[1 + M*exp(-A*Ci/1000)], where
    M = 4403, A = 126, and where Ci is the hourly average ``O3`` mixing
    ratio in units of ppb. If with the default min data capture
    criterion there are less than 9 valid hourly values in the 12-hour
    window, the daily value is considered missing. When there exist 75%
    or greater min data capture in the daily 12-hour window, the scaling
    by fractional min data capture (ntotal/nvalid) will be utilized.

    :param ser: a series containing the values
    :param ref: reference series
    :param mtype: aggregation type
    :param metadata: metadata needed for some statistics
    :param seasons: season names
    :param min_data_capture: a fractional coverage value

    :return: A list of dictionaries containing the processed data
    """
    hour = ser.index.hour
    day = (hour >= 8) & (hour <= 19)
    hourref = ref.index.hour
    dayref = (hourref >= 8) & (hourref <= 19)
    return stat_processor_1("w126", ser[day], ref[dayref], mtype, seasons,
                            func=w126stat, how="sum", minfrac=min_data_capture)


@register("w126_24h")
def w126_24h(ser, ref, mtype, metadata, seasons, min_data_capture):
    """Calculate the W126 metric over the 24 hour period of a day.

    Daily W126 index over the 24 hour period of a day. See ``w126`` for
    definition.

    :param ser: a series containing the values
    :param ref: reference series
    :param mtype: aggregation type
    :param metadata: metadata needed for some statistics
    :param seasons: season names
    :param min_data_capture: a fractional coverage value

    :return: A list of dictionaries containing the processed data
    """
    return stat_processor_1("w126_24h", ser, ref, mtype, seasons,
                            func=w126stat, how="sum", minfrac=min_data_capture)


@register("nvgt050")
def nvgt050(ser, ref, mtype, metadata, seasons, min_data_capture):
    """Calculate the NVGT050 metric.

    Calculate the number of days with exceedance of daily max. 8-hour
    values above 50 ppb. Not to be used with ``mtype = "daily"``.

    :param ser: a series containing the values
    :param ref: reference series
    :param mtype: aggregation type
    :param metadata: metadata needed for some statistics
    :param seasons: season names
    :param min_data_capture: a fractional coverage value

    :return: A list of dictionaries containing the processed data
    """
    new_ser = dma8epax(ser, ref, "daily", metadata, seasons,
                       min_data_capture)[0]["ser"]
    new_ref = ref.resample("D").mean()
    return stat_processor_1("nvgt050", new_ser, new_ref, mtype, seasons,
                            func=threshold_processor, minfrac=min_data_capture,
                            threshold=50.)


@register("nvgt060")
def nvgt060(ser, ref, mtype, metadata, seasons, min_data_capture):
    """Calculate the NVGT060 metric.

    Calculate the number of days with exceedance of daily max. 8-hour
    values above 60 ppb. Not to be used with ``mtype = "daily"``.

    :param ser: a series containing the values
    :param ref: reference series
    :param mtype: aggregation type
    :param metadata: metadata needed for some statistics
    :param seasons: season names
    :param min_data_capture: a fractional coverage value

    :return: A list of dictionaries containing the processed data
    """
    new_ser = dma8epax(ser, ref, "daily", metadata, seasons,
                       min_data_capture)[0]["ser"]
    new_ref = ref.resample("D").mean()
    return stat_processor_1("nvgt060", new_ser, new_ref, mtype, seasons,
                            func=threshold_processor, minfrac=min_data_capture,
                            threshold=60.)


@register("nvgt070")
def nvgt070(ser, ref, mtype, metadata, seasons, min_data_capture):
    """Calculate the NVGT070 metric.

    Calculate the number of days with exceedance of daily max. 8-hour
    values above 70 ppb. Not to be used with ``mtype = "daily"``.

    :param ser: a series containing the values
    :param ref: reference series
    :param mtype: aggregation type
    :param metadata: metadata needed for some statistics
    :param seasons: season names
    :param min_data_capture: a fractional coverage value

    :return: A list of dictionaries containing the processed data
    """
    new_ser = dma8epax(ser, ref, "daily", metadata, seasons,
                       min_data_capture)[0]["ser"]
    new_ref = ref.resample("D").mean()
    return stat_processor_1("nvgt070", new_ser, new_ref, mtype, seasons,
                            func=threshold_processor, minfrac=min_data_capture,
                            threshold=70.)


@register("nvgt080")
def nvgt080(ser, ref, mtype, metadata, seasons, min_data_capture):
    """Calculate the NVGT080 metric.

    Calculate the number of days with exceedance of daily max. 8-hour
    values above 80 ppb. Not to be used with ``mtype = "daily"``.

    :param ser: a series containing the values
    :param ref: reference series
    :param mtype: aggregation type
    :param metadata: metadata needed for some statistics
    :param seasons: season names
    :param min_data_capture: a fractional coverage value

    :return: A list of dictionaries containing the processed data
    """
    new_ser = dma8epax(ser, ref, "daily", metadata, seasons,
                       min_data_capture)[0]["ser"]
    new_ref = ref.resample("D").mean()
    return stat_processor_1("nvgt080", new_ser, new_ref, mtype, seasons,
                            func=threshold_processor, minfrac=min_data_capture,
                            threshold=80.)


@register("nvgt090")
def nvgt090(ser, ref, mtype, metadata, seasons, min_data_capture):
    """Calculate the NVGT090 metric.

    Calculate the number of days with exceedance of daily max. 1-hour
    values above 90 ppb. Note that in contrast to ``nvgt050``,
    ``nvgt060``, ``nvgt070`` and ``nvgt080``, this metric uses daily
    1-hourly maximum values as a basis. Not to be used with
    ``mtype = "daily"``.

    :param ser: a series containing the values
    :param ref: reference series
    :param mtype: aggregation type
    :param metadata: metadata needed for some statistics
    :param seasons: season names
    :param min_data_capture: a fractional coverage value

    :return: A list of dictionaries containing the processed data
    """
    new_ser = max1h_values(ser, ref, "daily", metadata, seasons,
                           min_data_capture)[0]["ser"]
    new_ref = ref.resample("D").mean()
    return stat_processor_1("nvgt090", new_ser, new_ref, mtype, seasons,
                            func=threshold_processor, minfrac=min_data_capture,
                            threshold=90.)


@register("nvgt100")
def nvgt100(ser, ref, mtype, metadata, seasons, min_data_capture):
    """Calculate the NVGT100 metric.

    Calculate the number of days with exceedance of daily max. 1-hour
    values above 100 ppb. Note that in contrast to ``nvgt050``,
    ``nvgt060``, ``nvgt070`` and ``nvgt080``, this metric uses daily
    1-hourly maximum values as a basis. Not to be used with
    ``mtype = "daily"``.

    :param ser: a series containing the values
    :param ref: reference series
    :param mtype: aggregation type
    :param metadata: metadata needed for some statistics
    :param seasons: season names
    :param min_data_capture: a fractional coverage value

    :return: A list of dictionaries containing the processed data
    """
    new_ser = max1h_values(ser, ref, "daily", metadata, seasons,
                           min_data_capture)[0]["ser"]
    new_ref = ref.resample("D").mean()
    return stat_processor_1("nvgt100", new_ser, new_ref, mtype, seasons,
                            func=threshold_processor, minfrac=min_data_capture,
                            threshold=100.)


@register("nvgt120")
def nvgt120(ser, ref, mtype, metadata, seasons, min_data_capture):
    """Calculate the NVGT120 metric.

    Calculate the number of days with exceedance of daily max. 1-hour
    values above 120 ppb. Note that in contrast to ``nvgt050``,
    ``nvgt060``, ``nvgt070`` and ``nvgt080``, this metric uses daily
    1-hourly maximum values as a basis. Not to be used with
    ``mtype = "daily"``.

    :param ser: a series containing the values
    :param ref: reference series
    :param mtype: aggregation type
    :param metadata: metadata needed for some statistics
    :param seasons: season names
    :param min_data_capture: a fractional coverage value

    :return: A list of dictionaries containing the processed data
    """
    new_ser = max1h_values(ser, ref, "daily", metadata, seasons,
                           min_data_capture)[0]["ser"]
    new_ref = ref.resample("D").mean()
    return stat_processor_1("nvgt120", new_ser, new_ref, mtype, seasons,
                            func=threshold_processor, minfrac=min_data_capture,
                            threshold=120.)


@register("nvgtall")
def nvgtall(ser, ref, mtype, metadata, seasons, min_data_capture):
    """Calculate all nvgt metrics together.

    :param ser: a series containing the values
    :param ref: reference series
    :param mtype: aggregation type
    :param metadata: metadata needed for some statistics
    :param seasons: season names
    :param min_data_capture: a fractional coverage value

    :return: A list of dictionaries containing the processed data
    """
    return [
        res for func in (nvgt050, nvgt060, nvgt080, nvgt090, nvgt100, nvgt120)
        for res in func(ser, ref, mtype, metadata, seasons, min_data_capture)
    ]


@register("osdma8")
def osdma8(ser, ref, mtype, metadata, seasons, min_data_capture):
    """ Calculate osdma8 statistic

    :param ser: a series containing the values
    :param ref: reference series
    :param mtype: aggregation type
    :param metadata: metadata needed for some statistics
    :param seasons: season names
    :param min_data_capture: a fractional coverage value

    :return: A list of dictionaries containing the processed data
    """
    monthly_means = stat_processor_1("mean", ser, ref, "monthly", seasons, mincount=10)

    tmp = monthly_means[0]["ser"]
    reference_index = pd.date_range(
        start=ref.index.min().strftime("%Y-%m-%d"),
        end=ref.index.max().strftime("%Y-%m-%d"),
        freq="MS"
    )
    tmp = tmp.reindex(reference_index)
    rolling_mean = tmp.rolling(window=6).mean()
    # rolling mean should be indexed by start month
    rolling_mean = rolling_mean.shift(-5)
    rolling_mean = rolling_mean[rolling_mean.index.month <= 10]

    rolling_mean_grouped = rolling_mean.groupby(rolling_mean.index.year)

    res = rolling_mean_grouped.agg(["max"])
    res.index = pd.to_datetime(res.index, format="%Y")

    res = res["max"]
    res = res.iloc[:-1]
    return [{"name": "osdma8", "ser": res, "ref": ref}]
